/* Copyright 2009 DJ Delorie <dj@delorie.com> */
/* Distributed under the terms of the GNU GPL, version 3. */

/* Reads in a 2880 DPI PBM bitmap (specified on the command line) and
   produces output.esc, which can be sent directly to an Epson ESC/P2
   high resolution color printer. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include "escp2.h"

typedef unsigned char byte;

int opt_draft = 0;
int opt_invert = 0;
int opt_mirror = 0;

int color_black = 1;
int color_cyan = 0;
int color_magenta = 0;
int color_yellow = 0;
int color_ltcyan = 0;
int color_ltmagenta = 0;

typedef struct {
  int width;
  int rwidth;
  int height;
  byte **pix;
} PBM;

int pbm_width;
int row_width;
unsigned char *hits;

#define mm(x) ((x)/25.4)

struct {
  int ink;
  int mask;
  int xofs;
  int yofs;
} colors[] = {
  /* color  mask xofs yofs */
  { CYAN,       1,   0, 14 },
  { MAGENTA,    2,   0,  0 },
  { YELLOW,     4,   0,  0 },
  { BLACK,      8,   0, 14 },
  { LTCYAN,    16,   0,  0 },
  { LTMAGENTA, 32,   0,  0 }
};
enum {
  C_CYAN,
  C_MAGENTA,
  C_YELLOW,
  C_BLACK,
  C_LTCYAN,
  C_LTMAGENTA
};

struct {
  char *name;
  double width;
  double height;
} paper_sizes[] = {
  { "letter", 8.5, 11.0 },
  { "halfletter", 5.5, 8.5 },
  { "hl", 5.5, 8.5 },
  { "half", 5.5, 8.5 },
  { "a4", mm(210), mm(297) },
  { "a5", mm(148), mm(210) },
  { "a6", mm(105), mm(148) }
};
#define NUM_PAPER_SIZES (sizeof(paper_sizes)/sizeof(paper_sizes[0]))

int
read_pbm_int (FILE *f)
{
  int c, rv = 0;
  while ((c = fgetc (f)) != EOF)
    {
      if (c == '#')
	do { c = fgetc (f); } while (c != '\n');
      if (isdigit (c))
	{
	  rv = c - '0';
	  while ((c = fgetc(f)) != EOF)
	    {
	      if (!isdigit (c))
		return rv;
	      rv *= 10;
	      rv += c - '0';
	    }
	  return rv;
	}
    }
}

PBM *
load_pbm (char *filename)
{
  int w, h, d;
  int i, x, y;
  FILE *f;
  PBM *pbm;

  f = fopen(filename, "rb");
  if (!f)
    {
      perror(filename);
      exit(1);
    }
  if (fgetc (f) != 'P'
      || fgetc (f) != '4')
    {
      printf("%s not a binary PBM file\n", filename);
      exit(1);
    }
  w = read_pbm_int (f);
  h = read_pbm_int (f);
  pbm = (PBM *)malloc (sizeof(PBM));
  pbm->width = w;
  pbm->height = h;
  pbm->pix = (byte **)malloc(w * sizeof (byte *));
  for (x=0; x<w; x++)
    pbm->pix[x] = (byte *)malloc(h);
  for (y=0; y<h; y++)
    for (x=0; x<(w+7)/8; x++)
      pbm->pix[x][y] = fgetc(f);
  fclose(f);
  return pbm;
}

static PBM *pbm;
int xoff = 0, yoff = 0;
int cur_v = 0;

static int
get_pixel (int x, int y, int c)
{
#ifdef RAINBOW_TEST
  switch (c)
    {
      /* 1440 pixels / 4 = 360,720,1080,1440 */
    case C_CYAN:
      return ( 144 < y && y <  432) ? 1 : 0;
    case C_MAGENTA:
      return ( 432 < y && y <  720) ? 1 : 0;
    case C_YELLOW:
      return ( 720 < y && y < 1008) ? 1 : 0;
    case C_BLACK:
      return (1008 < y && y < 1296) ? 1 : 0;
    default:
      return 0;
    }
#endif

  if (x < 0 || x >= pbm->width)
    return 0;
  if (y < 0 || y >= pbm->height)
    return 0;
  if (opt_mirror)
    x = pbm->width - x;
  if (pbm->pix[x/8][y] & (0x80 >> (x%8)))
    return 1-opt_invert;
  return opt_invert;
}

putint4(int i)
{
  putchar(i);
  putchar(i >> 8);
  putchar(i >> 16);
  putchar(i >> 24);
}

putint2(int i)
{
  putchar(i);
  putchar(i >> 8);
}

void
gotovh (int v, int h)
{
  int dv = (v + yoff) - cur_v;
  cur_v = (v + yoff);

  escp2_advance_row (dv);
  escp2_hpos (h + xoff);
}

static unsigned char *rbytes = 0;

void
send_row (int color, int r, int o)
{
  int x;

  if (r < 0 || r >= pbm->height
      || (hits[r*4+o] & colors[color].mask))
    {
      int c = row_width;
      while (c--)
	escp2_dot (0);
      return;
    }

  hits[r*4+o] |= colors[color].mask;

  r += colors[color].yofs;
  o += colors[color].xofs;

  for (x=0; x<row_width; x++)
    if (get_pixel (x*4+o, r, color))
      escp2_dot (1);
    else
      escp2_dot (0);
}

static void
do_band (int color, int top_row, int weave_idx)
{
  int i;

  gotovh (top_row, weave_idx);
  escp2_band (colors[color].ink, 2, row_width, 90);

  for (i=0; i<90; i++)
    {
      int r = top_row + 32 * i;
      send_row (color, r, weave_idx);
      escp2_end_row ();
    }
  escp2_end_band ();
}

#define STEP 17
#define WEAVE(x) ((x+x/5)%4)

static int
draft_include (int w, int r)
{
  switch (opt_draft)
    {
    case 0:
      return 1;
    case 1:
      return ((w % 2) == 0) && ((r % 2) == 0);
    case 2:
      return (w == 0) && ((r % 4) == 0);
    }
}

void
graphics ()
{
  int top_row;
  int block_idx;
  int pix_idx;
  int weave_idx, weave_cnt;

  /* ink jets are 32 rows apart, and print every 4 pixels across.  So,
     the ink head is 2880 pixels tall.  */

  top_row = -89 * 32;
  while (top_row < pbm->height)
    {
      weave_cnt ++;
      weave_idx = WEAVE(weave_cnt);
      if (draft_include (weave_idx, top_row))
	{
	  if (color_black)
	    do_band (C_BLACK, top_row, weave_idx);
	  if (color_cyan)
	    do_band (C_CYAN, top_row, weave_idx);
	  if (color_yellow)
	    do_band (C_YELLOW, top_row, weave_idx);
	  if (color_magenta)
	    do_band (C_MAGENTA, top_row, weave_idx);
	  if (color_ltcyan)
	    do_band (C_LTCYAN, top_row, weave_idx);
	  if (color_ltmagenta)
	    do_band (C_LTMAGENTA, top_row, weave_idx);
	}
      top_row += STEP;
    }
}

static void
usage ()
{
  int i;
  printf("Usage: pcm2escp2 [options] <file.pbm>\n");
  printf("Output is in output.esc\n");
  printf("-o <file>   set output file (default: output.esc)\n");
  printf("-d          slow draft mode (best quality on plain paper)\n");
  printf("-D          fast draft mode\n");
  printf("-i          invert pixels\n");
  printf("-m          mirror image\n");
  printf("-c <cmyk>   select inks [cmykCMrgb]+ (default: k)\n");
  printf("-x <xoff>   Offset from left edge of paper, in inches (default: 0)\n");
  printf("-y <yoff>   Offset from top edge of paper, in inches (min/default: 1)\n");

  printf("-p <paper>  paper size (centers)\n");
  for (i=0; i<NUM_PAPER_SIZES; i++)
    printf("\t%-12s %.1f x %.1f\n", paper_sizes[i].name,
	   paper_sizes[i].width, paper_sizes[i].height);

  exit(1);
}

int
main(int argc, char **argv)
{
  char optch;
  int paper = -1;
  int i;
  int xoff_set=0, yoff_set=0;

  while ((optch = getopt (argc, argv, "dDimp:c:x:y:o:")) != -1)
    {
      switch (optch)
	{
	case 'd':
	  opt_draft = 1;
	  break;

	case 'D':
	  opt_draft = 2;
	  break;

	case 'i':
	  opt_invert = 1;
	  break;

	case 'm':
	  opt_mirror = 1;
	  break;

	case 'p':
	  printf ("paper: `%s'\n", optarg);
	  for (i=0; i<NUM_PAPER_SIZES; i++)
	    if (strcasecmp (paper_sizes[i].name, optarg) == 0)
	      {
		paper = i;
		break;
	      }
	  if (paper == -1)
	    {
	      fprintf(stderr, "Unknown paper size: `%s'\n", optarg);
	      exit(1);
	    }
	  break;

	case 'c':
	  color_black = 0;
	  for (i=0; optarg[i]; i++)
	    switch (optarg[i])
	      {
	      case 'c': color_cyan = 1; break;
	      case 'm': color_magenta = 1; break;
	      case 'C': color_ltcyan = 1; break;
	      case 'M': color_ltmagenta = 1; break;
	      case 'y': color_yellow = 1; break;
	      case 'k': color_black = 1; break;
	      case 'r': color_magenta = color_yellow = 1; break;
	      case 'g': color_yellow = color_cyan = 1; break;
	      case 'b': color_cyan = color_magenta = 1; break;
	      }
	  break;

	case 'x':
	  xoff = strtod(optarg, 0) * 2880;
	  xoff_set = 1;
	  break;

	case 'y':
	  yoff = strtod(optarg, 0) * 2880;
	  yoff_set = 1;
	  break;

	case 'o':
	  escp2_set_output (optarg);
	  break;

	case '?':
	  usage ();
	}
    }

  if (!argv[optind])
    usage ();

  pbm = load_pbm (argv[optind]);

  hits = (unsigned char *) malloc (pbm->height * 4);

  pbm_width = 4 * ((pbm->width + 3) / 4);
  row_width = pbm->width / 4;

  if (paper != -1)
    {
      if (! xoff_set)
	{
	  xoff = paper_sizes[paper].width  * 2880 / 2 - pbm->width / 2;
	  if (xoff < 0)
	    xoff = 0;
	}
      if (! yoff_set)
	{
	  yoff = paper_sizes[paper].height * 2880 / 2 - pbm->height / 2;
	  if (yoff < 0)
	    yoff = 0;
	}
    }

  printf("color: C %d M %d Y %d K %d LC %d LM %d xoff: %g yoff: %g\n",
	 color_cyan, color_magenta, color_yellow, color_black, color_ltcyan, color_ltmagenta,
	 xoff/2880.0, yoff/2880.0);

  escp2_init (8.5, 11, opt_draft);
  graphics ();
  escp2_done ();
  
  exit(0);
}
